<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform" version="1.0">

	<xsl:output method="text"/>
	
    <xsl:param name="service" select="services/service[1]/@name"/>
    <xsl:param name="e4x" select="false()"/>
    
    <xsl:template match="/">
        <xsl:apply-templates select="services/service[@name=$service][1]"/>
    </xsl:template>
    
    <xsl:template match="service">
        <xsl:variable name="service-name">
        <xsl:call-template name="xml-name-to-javascript-name">
            <xsl:with-param name="name" select="@name"/>
        </xsl:call-template>
    </xsl:variable>
//  Example stubs for <xsl:value-of select="$service-name"/> operations.  This function is not intended 
//  to be called, but rather as a source for copy-and-paste development.

function stubs()
{
<xsl:for-each select="operations/operation">
    <xsl:variable name="name">
        <xsl:call-template name="xml-name-to-javascript-name">
            <xsl:with-param name="name" select="@name"/>
        </xsl:call-template>
    </xsl:variable>    // <xsl:value-of select="$name"/> operation
    try {
        <xsl:for-each select="signature/returns/param"><xsl:call-template name="return-type"/></xsl:for-each>
        <xsl:text> </xsl:text>
        <xsl:value-of select="$name"/>Return = <xsl:value-of select="$service-name"/>.<xsl:value-of select="$name"/>(<xsl:call-template name="parameters"><xsl:with-param name="prefix" select="'param_'"/></xsl:call-template>);
    } catch (e) {
        // fault handling
    }

</xsl:for-each>}
stubs.visible = false;

var <xsl:value-of select="$service-name"/> = new WebService("<xsl:for-each select="/services/service[@name=$service-name]">
            <xsl:sort select="@type = 'SOAP12'" order="descending"/>
            <xsl:sort select="@type = 'SOAP11'" order="descending"/>
            <xsl:sort select="@address" order="ascending"/>
            <xsl:if test="position() = 1"><xsl:value-of select="@endpoint"/></xsl:if>
        </xsl:for-each>");

<xsl:for-each select="operations/operation">
    <xsl:variable name="name">
    <xsl:call-template name="xml-name-to-javascript-name">
        <xsl:with-param name="name" select="@name"/>
    </xsl:call-template>
</xsl:variable>
<xsl:value-of select="$service-name"/>.<xsl:value-of select="$name"/> =
    function <xsl:value-of select="$name"/>(<xsl:call-template name="parameters"/>)
    {<xsl:call-template name="build-arrays"/>
        this._options = new Array();
        var isAsync = (typeof(this.<xsl:value-of select="$name"/>.callback) == 'function');
        var request = 
            <xsl:call-template name="payload"/><xsl:text>
        </xsl:text>
        <xsl:choose>
            <xsl:when test="wsaw/@action">this._options.useWSA = true;
        this._options.action = "<xsl:value-of select="wsaw/@action"/>";</xsl:when>
            <xsl:when test="binding-details/@soapaction">this._options.action = "<xsl:value-of select="binding-details/@soapaction"/>";</xsl:when>
        </xsl:choose>
        if (isAsync) {
            try {
                this._call(
                    request,
                    function(thisRequest, callbacks) {
                        if (thisRequest.error != null) {
                            callbacks[1](thisRequest.error);
                        } else {
                            var response = thisRequest.responseXML;
                            <xsl:choose>
                                <xsl:when test="signature/returns/@wrapper-element != '' or signature/returns/param/@simple = 'yes'">
                                    <xsl:apply-templates select="signature/returns/param" mode="convert"/>
                                </xsl:when>
                                <xsl:otherwise>
                                    <xsl:call-template name="default-convert"/>
                                </xsl:otherwise>
                            </xsl:choose>                            callbacks[0](resultValue);
                        }
                    },
                    new Array(this.<xsl:value-of select="$name"/>.callback, this.<xsl:value-of select="$name"/>.onError)
                );
            } catch (e) {
                error = new WebServiceError();
                if (typeof(e) == "string") {
                    error.code = "WSRequest: 0 (internal error)";
                    error.reason = e;
                    error.details = "";
                } else {
                    error.code = "WSRequest: " + e.number;
                    error.reason = e.description;
                    error.details = e.number;
                }
                this.<xsl:value-of select="$name"/>.onError(error);
            }
        } else {
            try {
                            var response = this._call(request);
                            <xsl:choose>
                                <xsl:when test="signature/returns/@wrapper-element != '' or signature/returns/param/@simple = 'yes'">
                                    <xsl:apply-templates select="signature/returns/param" mode="convert"/>
                                </xsl:when>
                                <xsl:otherwise>
                                    <xsl:call-template name="default-convert"/>
                                </xsl:otherwise>
                            </xsl:choose>                            return resultValue;
            } catch (e) {
                if (typeof(e) == "string") throw(e);
                if (e.message) throw(e.message);
                throw (e.reason);
            }
        }
    }
    
</xsl:for-each>

// WebService object.
function WebService(endpointName)
{
    this.readyState = 0;
    this.onreadystatechange = null;
    
    //public accessors for manually intervening in setting the address (e.g. supporting tcpmon)
    this.getAddress = function (endpointName)
    {
        return this._endpointDetails[endpointName].address;
    }
    
    this.setAddress = function (endpointName, address)
    {
        this._endpointDetails[endpointName].address = address;
    }

    // private helper functions
    this._getWSRequest = function()
    {
        var wsrequest;
        try {
            wsrequest = new WSRequest();
        } catch(e) {
            try {
                wsrequest = new ActiveXObject("WSRequest");
            } catch(e) {
                try {
                    wsrequest = new SOAPHttpRequest();
                    netscape.security.PrivilegeManager.enablePrivilege("UniversalXPConnect");
                } catch (e) {
                    try {
                        alert("WSRequest object not defined!");
                    } catch (e) {
                        print("WSRequest object not defined!");
                    }
                }
            }
        }
        return wsrequest;
    }
        
    this._endpointDetails = 
        {<xsl:for-each select="/services/service[@name=$service-name]">
            <xsl:sort select="@type = 'SOAP12'" order="descending"/>
            <xsl:sort select="@type = 'SOAP11'" order="descending"/>
            <xsl:sort select="@address" order="ascending"/>
            "<xsl:value-of select="@endpoint"/>": {
                "type" : "<xsl:value-of select="@type"/>",
                "address" : "<xsl:value-of select="@address"/>"
            }<xsl:if test="position() &lt; last()">,</xsl:if></xsl:for-each>
    };
    this.endpoint = endpointName;



    this._call = function (reqContent, callback, userdata)
    {
        var details = this._endpointDetails[this.endpoint];
        var oldAddress = details.address;
        if (details.type == 'SOAP12') this._options.useSOAP = 1.2;
        else if (details.type == 'SOAP11') this._options.useSOAP = 1.1;
        else if (details.type == 'HTTP') {
            this._options.useSOAP = false;
            var action = this._options.action;
            var opName;
                 var index = action.indexOf("urn:");
                 if (index == 0) {
                      opName = action.substr(4);
                      details.address = oldAddress + "/" + opName;
                 }
        }

        var isAsync = (typeof(callback) == 'function');
        
        var thisRequest = this._getWSRequest();
        if (isAsync) {
            thisRequest._userdata = userdata;
            thisRequest.onreadystatechange = 
                function() {
                    if (thisRequest.readyState == 4) {
                        callback(thisRequest, userdata);
                    }
                }
        }
        thisRequest.open(this._options, details.address, isAsync);
        thisRequest.send(reqContent);
        details.address = oldAddress;
        if (isAsync) {
            return "";
        } else {
            try {
                var resultContent = thisRequest.responseText;
                if (resultContent == "") { 
                    throw ("no response");
                }
                var resultXML = thisRequest.responseXML;
            } catch (e) {
                throw(e);
            }
            return resultXML;
        }
    }				
}
WebService.visible = false;
        
<xsl:if test="operations/operation//param[@type-namespace = 'http://www.w3.org/2001/XMLSchema' and
    ( @type = 'date' or
      @type = 'dateTime' or
      @type = 'time'
    )]">// library function for parsing xs:date, xs:time, and xs:dateTime types into Date objects.
function xs_dateTime_to_date(dateTime)
{
    var buffer = dateTime;
    var p = 0; // pointer to current parse location in buffer.
    
    // parse date, if there is one.
    if (buffer.substr(p,1) == '-')
    {
        var era = -1;
        p++;
    } else {
        var era = 1;
    }
    
    if (buffer.charAt(p+2) != ':')
    {
        var year = era * buffer.substr(p,4);
        p += 5;
        var month = buffer.substr(p,2);
        p += 3;
        var day = buffer.substr(p,2);
        p += 3;
    } else {
        var year = 1970;
        var month = 1;
        var day = 1;
    }
    
    // parse time, if there is one
    if (buffer.charAt(p) != '+' &amp;&amp; buffer.charAt(p) != '-')
    {
        var hour = buffer.substr(p,2);
        p += 3;
        var minute = buffer.substr(p,2);
        p += 3;
        var second = buffer.substr(p,2);
        p += 2;
        if (buffer.charAt(p) == '.')
        {
            var millisecond = parseFloat(buffer.substr(p))*1000;
            // Note that JS fractional seconds are significant to 3 places - xs:time is significant to more - 
            // though implementations are only required to carry 3 places.
            p++;
            while (buffer.charCodeAt(p) >= 48 &amp;&amp; buffer.charCodeAt(p) &lt;= 57) p++;
        } else {
            var millisecond = 0;
        }
    } else {
        var hour = 0;
        var minute = 0;
        var second = 0;
        var millisecond = 0;
    }
    
    var tzhour = 0;
    var tzminute = 0;
    // parse time zone
    if (buffer.charAt(p) != 'Z' &amp;&amp; buffer.charAt(p) != '') {
        var sign = (buffer.charAt(p) == '-' ? -1 : +1);
        p++;
        var tzhour = sign * buffer.substr(p,2);
        p += 3;
        var tzminute = sign * buffer.substr(p,2);
    }
    
    var thisDate = new Date();
    thisDate.setUTCFullYear(year);
    thisDate.setUTCMonth(month-1);
    thisDate.setUTCDate(day);
    thisDate.setUTCHours(hour);
    thisDate.setUTCMinutes(minute);
    thisDate.setUTCSeconds(second);
    thisDate.setUTCMilliseconds(millisecond);    
    thisDate.setUTCHours(thisDate.getUTCHours() - tzhour);
    thisDate.setUTCMinutes(thisDate.getUTCMinutes() - tzminute);
    return thisDate;
}
xs_dateTime_to_date.visible = false;</xsl:if>
</xsl:template>
    
    <!-- Non-RPC wrapped -->
    <xsl:template match="param[@token='#raw']" mode="convert" priority="2">
                            var resultValue = <xsl:call-template name="return-type"/> response;
    </xsl:template>

    <xsl:template match="param[@type-namespace = 'http://www.w3.org/2001/XMLSchema' and
                              ( @type = 'string' or
                                @type = 'normalizedString' or
                                @type = 'token' or
                                @type = 'language' or
                                @type = 'Name' or
                                @type = 'NCName' or
                                @type = 'ID' or
                                @type = 'IDREF' or
                                @type = 'NMTOKEN' or
                                @type = 'ENTITY' or
                                @type = 'anyURI' or
                                @type = 'hexBinary' or
                                @type = 'base64Binary' or
                                @type = 'decimal' or
                                @type = 'NOTATION'
                              )]" mode="convert">
                            <xsl:call-template name="extracted-value"/>
                            var resultValue = <xsl:call-template name="return-type"/> extractedValue;
</xsl:template>

<xsl:template match="param[@type-namespace = 'http://www.w3.org/2001/XMLSchema' and
                              ( @type = 'integer' or
                                @type = 'nonPositiveInteger' or
                                @type = 'negativeInteger' or
                                @type = 'long' or
                                @type = 'int' or
                                @type = 'short' or
                                @type = 'byte' or
                                @type = 'nonNegativeInteger' or
                                @type = 'unsignedLong' or
                                @type = 'unsignedInt' or
                                @type = 'unsignedShort' or
                                @type = 'unsignedByte' or
                                @type = 'positiveInteger' or
                                @type = 'duration'  or
                                @type = 'gYear'  or
                                @type = 'gMonth'  or
                                @type = 'gDay' 
                              )]" mode="convert">
                            <xsl:call-template name="extracted-value"/>
                            var resultValue = <xsl:call-template name="return-type"/>
 parseInt(extractedValue);
</xsl:template>

<xsl:template match="param[@type-namespace = 'http://www.w3.org/2001/XMLSchema' and
                              ( @type = 'float' or
                                @type = 'double'
                              )]" mode="convert">
                            <xsl:call-template name="extracted-value"/>
                            var resultValue = <xsl:call-template name="return-type"/> parseFloat(extractedValue);
</xsl:template>

<xsl:template match="param[@type-namespace = 'http://www.w3.org/2001/XMLSchema' and
                              ( @type = 'boolean' )]" mode="convert">
                            <xsl:call-template name="extracted-value"/>
                            var resultValue = /* Boolean */ extractedValue == "true" || extractedValue == "1";
</xsl:template>

<!-- todo: add appropriate namespace declaration for QName input parameters too 
This looks broken - QName is copied from E4X  - what's the alternative in pure JS?
-->
<xsl:template match="param[@type-namespace = 'http://www.w3.org/2001/XMLSchema' and
                            ( @type = 'QName' )]" mode="convert">
                            <xsl:call-template name="extracted-value"/>
                            var resultValue = /* QName */ new QName(extractedValue, extractedValue.namespace(extractedValue.substr(0, extractedValue.indexOf(':'))));
</xsl:template>

<!-- todo: Do a better job than String for the following date types 
    <simple type="xs:gMonthDay"/>
    <simple type="xs:gYearMonth"/>
-->
<xsl:template match="param[@type-namespace = 'http://www.w3.org/2001/XMLSchema' and
                            ( @type = 'gMonthDay' or
                              @type = 'gYearMonth'
                            )]" mode="convert">
                            <xsl:call-template name="extracted-value"/>
                            var resultValue = <xsl:call-template name="return-type"/> extractedValue.toString();
</xsl:template>

<xsl:template match="param[@type-namespace = 'http://www.w3.org/2001/XMLSchema' and
                            ( @type = 'gMonthDay' or
                              @type = 'gYearMonth'
                            )]" mode="convert">
                            <xsl:call-template name="extracted-value"/>
                            var resultValue = /* Date */ extractedValue.split(' '));
</xsl:template>

<!-- todo: serialize Date objects into schema dates -->
<xsl:template match="param[@type-namespace = 'http://www.w3.org/2001/XMLSchema' and
                        ( @type = 'date' or
                          @type = 'dateTime' or
                          @type = 'time'
                        )]" mode="convert">
                            <xsl:call-template name="extracted-value"/>
                            var resultValue = /* Date */ xs_dateTime_to_date(extractedValue);
</xsl:template>

<!-- todo: convert input arrays into space-delimited lists -->
<xsl:template match="param[@type-namespace = 'http://www.w3.org/2001/XMLSchema' and
                            ( @type = 'ENTITIES' )]" mode="convert">
        <xsl:call-template name="extracted-value"/>
                            var resultValue = /* array of xs:ENTITY */ extractedValue.split(' '));
</xsl:template>
    
<xsl:template match="param[@type-namespace = 'http://www.w3.org/2001/XMLSchema' and
                            ( @type = 'IDREFS' )]" mode="convert">
                            <xsl:call-template name="extracted-value"/>
                            var resultValue = /* array of xs:IDREF */ extractedValue.split(' '));
</xsl:template>

<xsl:template match="param[@type-namespace = 'http://www.w3.org/2001/XMLSchema' and
                            ( @type = 'NMTOKENS' )]" mode="convert">
                            <xsl:call-template name="extracted-value"/>
                            var resultValue = /* array of xs:NMTOKEN */ extractedValue.split(' '));
</xsl:template>
    
<xsl:template match="param[@type-namespace = 'http://www.w3.org/2001/XMLSchema' and
                          ( @type = 'anyType' )]" mode="convert">
    <!-- FIXME MASHUP-25: &lt;return&gt; should not be in a namespace. -->
    <xsl:choose>
        <xsl:when test="$e4x">var resultValue = /* xs:anyType */ response["<xsl:value-of select="@name"/>"].*;
</xsl:when>
        <xsl:otherwise>var resultValue = /* xs:anyType */ response.documentElement.firstChild.firstChild;
</xsl:otherwise>
    </xsl:choose>
</xsl:template>
<xsl:template name="default-convert">
    <xsl:choose>
        <xsl:when test="$e4x">var resultValue = /* xs:anyType */ response;
</xsl:when>
        <xsl:otherwise>var resultValue = /* xs:anyType */ response.documentElement;
</xsl:otherwise>
    </xsl:choose>
</xsl:template>
    
<!-- Default template - no XSD simple type was found, return the whole payload as XML -->
<xsl:template match="param" mode="convert">
                            var resultValue = <xsl:call-template name="return-type"/> response;
</xsl:template>

<xsl:template name="extracted-value">
    <xsl:choose>
        <xsl:when test="$e4x">var ns = new Namespace("<xsl:value-of select="../@wrapper-element-ns"/>");
                            var extractedValue = response.ns::<xsl:value-of select="../@wrapper-element"/>["<xsl:value-of select="@name"/>"].toString();</xsl:when>
        <xsl:otherwise>var extractedValue = WSRequest.util._stringValue(response.documentElement);</xsl:otherwise>
    </xsl:choose>
</xsl:template>

<xsl:template name="parameters">
    <xsl:param name="prefix" select="'_'"/>
    <xsl:for-each select="signature/params/param">
        <xsl:choose>
            <xsl:when test="@token = '#any'">/* XMLList */ additionalParameters</xsl:when>
            <xsl:otherwise>/* <xsl:if test="@maxOccurs = 'unbounded' or @maxOccurs > 1">array of </xsl:if><xsl:value-of select="@type"/> */ <xsl:value-of select="$prefix"/><xsl:value-of select="@name"/></xsl:otherwise>
        </xsl:choose>
        <xsl:if test="position() != last()">, </xsl:if>
    </xsl:for-each>
</xsl:template>

<xsl:template name="payload">
    <xsl:choose>
        <xsl:when test="signature/params/@wrapper-element">'&lt;p:<xsl:value-of select="signature/params/@wrapper-element"/> xmlns:p="<xsl:value-of select="signature/params/@wrapper-element-ns"/>
            <xsl:text>"&gt;' +</xsl:text>
            <xsl:for-each select="signature/params/param">
            <xsl:choose>
                <xsl:when test="@token = '#any'">
                additionalParameters + '</xsl:when>
                <xsl:when test="@maxOccurs = 'unbounded' or @maxOccurs > 1">
                _array_of_<xsl:value-of select="@name"/> + </xsl:when>
                <xsl:when test="@type = 'anyType' and @type-namespace='http://www.w3.org/2001/XMLSchema'">
                '&lt;<xsl:value-of select="@name"/><xsl:if test="@namespace"> xmlns="<xsl:value-of select="@namespace"/>"</xsl:if>&gt;' + _<xsl:value-of select="@name"/> + '&lt;/<xsl:value-of select="@name"/>&gt;' +</xsl:when>
                <xsl:when test="(@type = 'string' or @type = 'normalizedString' or @type = 'anyURI') and @type-namespace='http://www.w3.org/2001/XMLSchema'">
                '&lt;<xsl:value-of select="@name"/><xsl:if test="@namespace"> xmlns="<xsl:value-of select="@namespace"/>"</xsl:if>&gt;' + (_<xsl:value-of select="@name"/>.indexOf('&lt;')>=0 || _<xsl:value-of select="@name"/>.indexOf('&amp;')>=0 ? '&lt;![CDATA[' : '') + _<xsl:value-of select="@name"/> + (_<xsl:value-of select="@name"/>.indexOf('&lt;')>=0 || _<xsl:value-of select="@name"/>.indexOf('&amp;')>=0 ? ']]&gt;' : '') + '&lt;/<xsl:value-of select="@name"/>&gt;' +</xsl:when>
                <xsl:otherwise>
                '&lt;<xsl:value-of select="@name"/><xsl:if test="@namespace"> xmlns="<xsl:value-of select="@namespace"/>"</xsl:if>&gt;' + _<xsl:value-of select="@name"/> + '&lt;/<xsl:value-of select="@name"/>&gt;' +</xsl:otherwise>
                </xsl:choose>
            </xsl:for-each>
            '&lt;/p:<xsl:value-of select="signature/params/@wrapper-element"/>&gt;' ;</xsl:when>
        <xsl:when test="signature/params/param[@type = 'anyType' and @type-namespace='http://www.w3.org/2001/XMLSchema']">_<xsl:value-of select="signature/params/param/@name"/>;</xsl:when>
        <xsl:otherwise>"&lt;<xsl:value-of select="@name"/>/&gt;";<!-- @@ bug - should be able to send an empty string... --></xsl:otherwise>
    </xsl:choose>
</xsl:template>

<xsl:template name="build-arrays">
    <xsl:for-each select="signature/params/param[not(@token='#any')][@maxOccurs = 'unbounded' or @maxOccurs > 1]">
        <xsl:variable name="name">
            <xsl:call-template name="xml-name-to-javascript-name">
                <xsl:with-param name="name" select="@name"/>
            </xsl:call-template>
        </xsl:variable>
        var _array_of_<xsl:value-of select="$name"/> = '';
        for (var i=0; i &lt; _<xsl:value-of select="$name"/>.length; i++)
            _array_of_<xsl:value-of select="$name"/> += '&lt;<xsl:value-of select="$name"/>&gt;' + _<xsl:value-of select="$name"/>[i] + '&lt;/<xsl:value-of select="$name"/>&gt;';
    </xsl:for-each>
</xsl:template>

<xsl:template name="return-type">
    <xsl:text>/* </xsl:text>
    <xsl:if test="@maxOccurs = 'unbounded' or @maxOccurs > 1">array of </xsl:if>
    <xsl:call-template name="xml-name-to-javascript-name">
        <xsl:with-param name="name" select="@type"/>
    </xsl:call-template>
    <xsl:text>*/</xsl:text>
</xsl:template>

    <xsl:template name="xml-name-to-javascript-name">
        <xsl:param name="name"/>
        <xsl:value-of select="translate($name,'.-','__')"/>
    </xsl:template>

</xsl:stylesheet>